from typing import (
    Any,
    Dict,
)

from m3_ext.ui.fields.simple import (
    ExtCheckBox,
    ExtComboBox,
    ExtDateTimeField,
    ExtDisplayField,
    ExtNumberField,
    ExtStringField,
)
from m3_ext.ui.misc import (
    ExtDataStore,
)

from educommon.objectpack.ui import (
    BaseEditWindow,
)
from educommon.utils.date import (
    DatesSplitter,
)

from edu_rdm_integration.core.consts import (
    BATCH_SIZE,
    CHUNK_MAX_VALUE,
    SPLIT_BY_QUANTITY_MAX_VALUE,
)
from edu_rdm_integration.core.registry.ui import (
    BaseCreateCommandWindow,
)
from edu_rdm_integration.rdm_models.models import (
    RDMModelEnum,
)
from edu_rdm_integration.stages.collect_data.consts import (
    ALL_UNITS_IN_COMMAND,
)


class CreateCollectCommandWindow(BaseCreateCommandWindow):
    """Окно создания команды сбора данных модели РВД."""

    def _init_components(self):
        """Инициализация компонентов."""
        super()._init_components()

        model = ExtComboBox(
            name='model_id',
            label='Модель',
            display_field='model',
            anchor='100%',
            editable=False,
            trigger_action_all=True,
            allow_blank=False,
        )
        model.set_store(ExtDataStore((idx, key) for idx, key in enumerate(RDMModelEnum.get_model_enum_keys())))
        logs_period_started_at = ExtDateTimeField(
            name='logs_period_started_at',
            label='Начало периода',
            anchor='100%',
            allow_blank=False,
            client_id='logs_period_started_at',
        )
        logs_period_ended_at = ExtDateTimeField(
            name='logs_period_ended_at',
            label='Конец периода',
            anchor='100%',
            allow_blank=False,
            client_id='logs_period_ended_at',
        )
        split_by = ExtComboBox(
            name='split_by',
            display_field='split_by',
            label='Единица подпериода',
            anchor='100%',
            editable=False,
            trigger_action_all=True,
            client_id='split_by',
        )
        split_by.set_store(ExtDataStore(enumerate(DatesSplitter.get_split_by_modes())))
        split_by_quantity = ExtNumberField(
            name='split_by_quantity',
            label='Размер подпериода',
            allow_blank=False,
            allow_decimals=False,
            allow_negative=False,
            max_value=SPLIT_BY_QUANTITY_MAX_VALUE,
            anchor='100%',
            value=1,
        )
        split_mode = ExtComboBox(
            name='split_mode',
            display_field='split_mode',
            label='Режим разбиения на подпериоды',
            anchor='100%',
            editable=False,
            allow_blank=True,
            trigger_action_all=True,
            value=DatesSplitter.WW_MODE,
        )
        split_mode.set_store(
            ExtDataStore(enumerate(DatesSplitter.get_modes())),
        )
        batch_size = ExtNumberField(
            name='batch_size',
            label='Размер чанка',
            allow_blank=False,
            allow_decimals=False,
            allow_negative=False,
            anchor='100%',
            value=BATCH_SIZE,
            max_value=CHUNK_MAX_VALUE,
            client_id='batch_size',
        )
        by_institutes = ExtCheckBox(
            anchor='100%',
            label='Разбить по организациям',
            name='by_institutes',
        )
        institute_ids = ExtStringField(
            label='ID организаций (через запятую, например: 1,2,3 или оставить пустым для всех)',
            name='institute_ids',
            allow_blank=True,
            anchor='100%',
            max_length=100,
            client_id='institute_ids',
        )
        institute_count = ExtNumberField(
            name='institute_count',
            label='Кол-во организаций в одной команде',
            allow_decimals=False,
            anchor='100%',
            min_value=ALL_UNITS_IN_COMMAND,
            value=ALL_UNITS_IN_COMMAND,
            client_id='institute_count',
            allow_blank=False,
        )
        hint_text = ExtDisplayField(
            value=(
                f'Данные можно разбить или по "{batch_size.label}" или по "{split_by.label}"! '
                f'Если выбрать оба варианта, то будет выбрано разбиение по "{split_by.label}".'
            ),
            read_only=True,
            label_style='width: 0px',
            style={'text-align': 'center'},
        )
        just_or_hint_text = ExtDisplayField(
            label='или',
        )
        self.items_ = (
            model,
            logs_period_started_at,
            logs_period_ended_at,
            by_institutes,
            institute_ids,
            institute_count,
            hint_text,
            batch_size,
            just_or_hint_text,
            split_by,
            split_by_quantity,
            split_mode,
        )

    def set_params(self, params: Dict[str, Any]) -> None:
        """Устанавливает параметры окна."""
        super().set_params(params)

        self.template_globals = 'ui-js/collect-command-window.js'


class DetailCollectCommandWindow(BaseEditWindow):
    """Окно просмотра команды сбора данных модели РВД."""

    def set_params(self, params: Dict[str, Any]) -> None:
        """Устанавливает параметры окна."""
        super().set_params(params)

        self.height = 'auto'
        self.logs_link_field.value = params.get('log_url', '')

    def _init_components(self) -> None:
        """Инициализирует компоненты окна."""
        super()._init_components()

        self.model_field = ExtStringField(
            name='model_id',
            label='Модель',
            anchor='100%',
        )
        self.created_field = ExtDateTimeField(
            name='created',
            label='Дата создания',
            anchor='100%',
        )
        self.generation_id_field = ExtStringField(
            name='generation_id',
            label='Идентификатор генерации',
            anchor='100%',
        )
        self.task_id_field = ExtStringField(
            name='task_id',
            label='Идентификатор задачи',
            anchor='100%',
        )
        self.institute_ids_field = ExtStringField(
            name='institute_ids',
            label='Идентификаторы организаций',
            anchor='100%',
        )
        self.status_field = ExtStringField(
            name='stage.status.key',
            label='Статус сбора',
            anchor='100%',
        )
        self.started_at_field = ExtDateTimeField(
            name='stage.started_at',
            label='Время начала сбора',
            anchor='100%',
        )
        self.logs_link_field = ExtDisplayField(
            name='log_url',
            label='Ссылка на логи',
            anchor='100%',
        )
        self.logs_period_started_at_field = ExtDateTimeField(
            name='logs_period_started_at',
            label='Начало периода',
            anchor='100%',
        )
        self.logs_period_ended_at_field = ExtDateTimeField(
            name='logs_period_ended_at',
            label='Конец периода',
            anchor='100%',
        )

    def _do_layout(self) -> None:
        """Располагает компоненты окна."""
        super()._do_layout()

        self.form.items.extend(
            (
                self.model_field,
                self.created_field,
                self.generation_id_field,
                self.task_id_field,
                self.institute_ids_field,
                self.status_field,
                self.started_at_field,
                self.logs_link_field,
                self.logs_period_started_at_field,
                self.logs_period_ended_at_field,
            )
        )



from django.db.models import (
    CASCADE,
    BooleanField,
    OneToOneField,
    PositiveIntegerField,
)
from m3 import (
    json_encode,
)
from m3.db import (
    BaseObjectModel,
)

from edu_rdm_integration.pipelines.transfer.enums import (
    EntityLevelQueueTypeEnum,
)
from edu_rdm_integration.rdm_entities.models import (
    RDMEntityEnum,
)


class TransferredEntity(BaseObjectModel):
    """Сущность, по которой должен быть произведен сбор и экспорт данных."""

    entity = OneToOneField(
        to=RDMEntityEnum,
        verbose_name='Сущность',
        on_delete=CASCADE,
    )
    export_enabled = BooleanField(
        verbose_name='Включение экспорта для сущности',
        default=True,
    )
    queue_level = PositiveIntegerField(
        choices=EntityLevelQueueTypeEnum.get_choices(),
        default=EntityLevelQueueTypeEnum.BASE,
        verbose_name='Уровень очереди',
    )

    class Meta:
        db_table = 'rdm_transferred_entity'
        verbose_name = 'Сущность, по которой должен быть произведен сбор и экспорт данных'
        verbose_name_plural = 'Сущности, по которым должен быть произведен сбор и экспорт данных'

    @json_encode
    def no_export(self):
        """Формирует отображение признака отключения экспорта."""
        return 'Нет' if self.export_enabled else 'Да'



# Generated by Django 3.2.24 on 2025-06-28 14:22

from django.db import (
    migrations,
)


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('uploader_client', '0001_initial'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='UploaderClientLog',
                    fields=[],
                    options={
                        'proxy': True,
                        'indexes': [],
                        'constraints': [],
                    },
                    bases=('uploader_client.entry',),
                ),
            ],
            database_operations=[],
        ),
    ]



from django.conf import (
    settings,
)
from m3 import (
    OperationResult,
)
from m3_ext.ui.containers import (
    ExtGridCheckBoxSelModel,
)
from objectpack.actions import (
    BaseAction,
    BaseWindowAction,
    ObjectPack,
)
from objectpack.filters import (
    ColumnFilterEngine,
)

from edu_rdm_integration.core.helpers import (
    BaseTaskStarter,
    make_download_link,
)
from edu_rdm_integration.core.registry.ui import (
    CommandProgressListWindow,
    CommandQueueSelectWindow,
)


class BaseCommandProgressPack(ObjectPack):
    """Базовый пак прогресса выполнения команд сбора/экспорта данных."""

    list_window = CommandProgressListWindow
    can_delete = False

    select_related = ['task', 'stage__status']

    filter_engine_clz = ColumnFilterEngine

    def __init__(self):
        super().__init__()

        self.queue_select_win_action = QueueLevelSelectWinAction()

        self.actions.append(self.queue_select_win_action)

    def get_list_window_params(self, params, request, context):
        """Получает параметры окна списка."""
        params = super().get_list_window_params(params, request, context)

        params['queue_select_win_url'] = self.queue_select_win_action.get_absolute_url()

        return params

    def get_edit_window_params(self, params, request, context):
        """Возвращает словарь параметров, которые будут переданы окну редактирования."""
        params = super().get_edit_window_params(params, request, context)

        if not params['create_new']:
            params['read_only'] = True
            obj = params['object']
            params['log_url'] = make_download_link(obj.logs_link)

        return params

    def configure_grid(self, grid, *args, **kwargs):
        """Конфигурирует грид."""
        super().configure_grid(grid, *args, **kwargs)

        grid.sm = ExtGridCheckBoxSelModel()
        grid.top_bar.button_new.text = 'Сгенерировать команды'
        grid.top_bar.button_edit.text = 'Просмотр'
        grid.top_bar.button_edit.icon_cls = 'icon-application-view-detail'
        grid.context_menu_row.menuitem_edit.text = 'Просмотр'
        grid.context_menu_row.menuitem_edit.icon_cls = 'icon-application-view-detail'

    def extend_menu(self, menu):
        """Расширяет главное меню."""
        if settings.RDM_MENU_ITEM:
            return menu.SubMenu(
                'Администрирование',
                menu.SubMenu(
                    'Региональная витрина данных',
                    menu.Item(
                        self.title,
                        self.list_window_action,
                    ),
                    icon='menu-dicts-16',
                ),
            )


class BaseStartTaskAction(BaseAction):
    """Базовый экшн создания асинхронных задач для выгрузки РВД."""

    url: str = None

    task_starter: BaseTaskStarter = None

    def run(self, request, context):
        """Непосредственное исполнение запроса."""
        queue_level = getattr(context, 'queue_level', None)
        result = self.task_starter().run(command_ids=context.commands, queue_level=queue_level)  # noqa pylint: disable=not-callable

        return OperationResult(
            success=True,
            message=result,
        )


class QueueLevelSelectWinAction(BaseWindowAction):
    """Экшен окна выбора очереди для запуска ручной команды сбора/выгрузки."""

    def create_window(self):
        """Создание окна."""
        self.win = CommandQueueSelectWindow()



from educommon.utils.enums import (
    HashGostFunctionVersion,
)


REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA = 'regional_data_mart_integration_collecting_data'
REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA = 'regional_data_mart_integration_exporting_data'

# Формат даты. Используется для выгрузки
DATE_FORMAT = '%Y-%m-%d'

# Формат даты/времени. Для выгрузки не используется (в выгрузке ISO формат)
DATETIME_FORMAT = '%d.%m.%Y %H:%M:%S'

# Формат даты/времени. Для выгрузки
EXPORT_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'

LOGS_DELIMITER = '    '

HASH_ALGORITHM = HashGostFunctionVersion.GOST12_512

BATCH_SIZE = 5000

CHUNK_MAX_VALUE = 100000
"""Максимальное значение для размера чанка при сборе данных РВД.
Ограничение 100000 предотвращает чрезмерное потребление памяти и блокировку БД
при обработке больших объемов данных в одном батче."""

SPLIT_BY_QUANTITY_MAX_VALUE = 366
"""Максимальное значение для размера подпериода при временном разделении данных."""

ACADEMIC_YEAR = {
    'start_day': 1,
    'start_month': 9,
    'end_day': 31,
    'end_month': 8,
}

TASK_QUEUE_NAME = 'RDM'
FAST_TRANSFER_TASK_QUEUE_NAME = 'RDM_FAST'
LONG_TRANSFER_TASK_QUEUE_NAME = 'RDM_LONG'



# Generated by Django 3.2.23 on 2025-02-04 14:13

from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):
    dependencies = [
        ('rdm_collect_and_export_data', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='edurdmcollectdatacommandprogress',
            options={'verbose_name': 'Задача по сбору данных', 'verbose_name_plural': 'Задачи по сбору данных'},
        ),
        migrations.AlterModelOptions(
            name='edurdmexportdatacommandprogress',
            options={'verbose_name': 'Команда экспорта данных', 'verbose_name_plural': 'Команды экспорта данных'},
        ),
        migrations.AlterField(
            model_name='edurdmcollectdatacommandprogress',
            name='institute_ids',
            field=models.JSONField(blank=True, default=list, null=True, verbose_name='id учебного заведения'),
        ),
    ]



from typing import (
    Optional,
)

from django.core.exceptions import (
    FieldDoesNotExist,
)
from django.db.models import (
    Subquery,
)

from edu_rdm_integration.rdm_models.models import (
    RDMModelEnum,
)
from edu_rdm_integration.stages.export_data.models import (
    RDMExportingDataCommandProgress,
    RDMExportingDataStage,
    RDMExportingDataSubStage,
    RDMExportingDataSubStageAttachment,
    RDMExportingDataSubStageEntity,
)
from educommon.utils.seqtools import (
    make_chunks,
)

from .base import (
    BaseServiceOutdatedDataCleaner,
    ServiceFileCleaner,
)
from .consts import (
    OLD_RDM_MODEL,
    UNION_CHUNK_SIZE,
)


class ExportingDataSubStageCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка подэтапов выгрузки данных, которые не ссылаются ни на одну модель РВД.

    Подход:
        - Проходим по всем моделям, зарегистрированным в RDMModelEnum, и собираем
            значения полей exporting_sub_stage_id (если модель содержит такое поле).
        - Объединяем запросы по моделям в UNION, получая набор валидных id.
        - Удаляем те подэтапы, id которых отсутствуют в полученном наборе.
    """
    model = RDMExportingDataSubStage

    def _get_valid_substage_ids_subquery(self) -> Optional[Subquery]:
        """Подзапрос, возвращающий все допустимые exporting_sub_stage_id из моделей, описанных в RDMModelEnum."""
        model_enum_values = RDMModelEnum.get_model_enum_values()
        all_model = [model_enum.model for model_enum in model_enum_values] + OLD_RDM_MODEL
        chunk_queries = []

        for enum_values_chunk in make_chunks(all_model, UNION_CHUNK_SIZE, is_list=True):
            qs_list = []
            for model_cls in enum_values_chunk:
                try:
                    model_cls._meta.get_field('exporting_sub_stage_id')
                except FieldDoesNotExist:
                    continue
                qs_list.append(model_cls.objects.values('exporting_sub_stage_id'))

            if qs_list:
                chunk_union = qs_list[0].union(*qs_list[1:])
                chunk_queries.append(chunk_union)

        if not chunk_queries:
            return

        # Объединяем все чанки в один общий UNION
        full_union = chunk_queries[0].union(*chunk_queries[1:])

        return Subquery(full_union.values('exporting_sub_stage_id'))

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных.

         Удаляем подэтапы, которых нет в объединённом наборе валидных exporting_sub_stage_id
        (т.е. подэтапы, не используемые ни одной моделью данных).
        """
        conditions = ''
        subquery = self._get_valid_substage_ids_subquery()
        if subquery:
           conditions = f"""
                NOT EXISTS (
                    SELECT exporting_sub_stage_id
                    FROM ({str(subquery.query)}) AS valid
                    WHERE valid.exporting_sub_stage_id = tbl.id
                )
            """

        return conditions


class ExportingDataStageCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка этапов выгрузки данных без подэтапов."""

    model = RDMExportingDataStage

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        sub_stage_table = ExportingDataSubStageCleaner.get_table_name()

        return self.get_orphan_reference_condition(sub_stage_table, 'stage_id')


class ExportingDataSubStageAttachmentCleaner(ServiceFileCleaner, BaseServiceOutdatedDataCleaner):
    """Очистка вложений подэтапов выгрузки данных."""

    model = RDMExportingDataSubStageAttachment

    REMOVE_OUTDATED_DATA_SQL = """
        WITH deleted_rows AS (
            DELETE FROM {table_name}
            WHERE id IN (
                WITH tbl AS (
                    SELECT *
                    FROM {table_name}
                    WHERE id >= {first_id}
                        AND id <= {last_id}
                )
                SELECT tbl.id
                FROM tbl
                WHERE {conditions}
            )
            RETURNING attachment AS file_path
        )
        SELECT file_path FROM deleted_rows;
    """

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        sub_stage_table = ExportingDataSubStageCleaner.get_table_name()
        conditions = [
            'exporting_data_sub_stage_id IS NULL',
            f'({self.get_status_condition(sub_stage_table, "id", "FINISHED", 7, "exporting_data_sub_stage_id")})',
            f'({self.get_status_condition(sub_stage_table, "id", "FAILED",30, "exporting_data_sub_stage_id")})',
            f'({self.get_orphan_reference_condition(sub_stage_table, "id", "exporting_data_sub_stage_id")})',
        ]

        return ' OR '.join(conditions)


class ExportingDataSubStageEntityCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка связей сущности и подэтапов выгрузки данных."""

    model = RDMExportingDataSubStageEntity

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        sub_stage_table = ExportingDataSubStageCleaner.get_table_name()
        conditions = [
            f'({self.get_status_condition(sub_stage_table, "id", "FINISHED", 7, "exporting_data_sub_stage_id")})',
            f'({self.get_status_condition(sub_stage_table, "id", "FAILED",30, "exporting_data_sub_stage_id")})',
            f'({self.get_orphan_reference_condition(sub_stage_table, "id", "exporting_data_sub_stage_id")})',
        ]

        return ' OR '.join(conditions)


class ExportingDataCommandProgressCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка устаревших хранящихся задач по экспорту данных."""

    model = RDMExportingDataCommandProgress

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        stage_table = ExportingDataStageCleaner.get_table_name()
        conditions = [
            'stage_id IS NULL',
            f'({self.get_status_condition(stage_table, "id", "FINISHED", 7, "stage_id")})',
            f'({self.get_status_condition(stage_table, "id", "FAILED",30, "stage_id")})',
            f'({self.get_orphan_reference_condition(stage_table, "id", "stage_id")})',
        ]

        return ' OR '.join(conditions)



from typing import (
    Optional,
    Type,
)

from educommon.integration_entities.entities import (
    BaseEntity,
)
from m3_db_utils.models import (
    ModelEnumValue,
    TitledModelEnum,
)


class RDMEntityEnum(TitledModelEnum):
    """Модель-перечисление сущностей выгрузки в Региональная витрина данных.

    Поля:
        entity - data-класс сущности;
        main_model_enum - значение модели-перечисления RegionalDataMartModelEnum основной модели РВД для формирования
            данных сущности. Обычно это модель идентификаторы записей которой соответствуют идентификаторам в записях
            сущности. У этих записей проставляется подэтап выгрузки данных;
        additional_model_enums - кортеж значений модели-перечисления RegionalDataMartModelEnum с дополнительными
            моделями РВД, которые участвуют в формировании записей сущностей. Они показывают, что перед запуском
            экспорта данных сущности по ним и основной модели должен быть запущен сбор данных.
    """

    class Meta:
        db_table = 'rdm_entity'
        extensible = True
        verbose_name = 'Модель-перечисление сущностей "Региональной витрины данных"'
        verbose_name_plural = 'Модели-перечисления сущностей "Региональной витрины данных"'

    @property
    def model_enums(self):
        """Возвращает модели, от которых зависит сущность."""
        value = self.model_enum_value

        return self.get_model_enums_from_value(value=value)

    @staticmethod
    def get_model_enums_from_value(value: ModelEnumValue):
        """Получение значений модели-перечисления моделей по значению модели-перечисления сущностей."""
        model_enums = [value.main_model_enum, *value.additional_model_enums]

        return model_enums

    @classmethod
    def get_entities_model_enums(
        cls,
        entity_enums: list[ModelEnumValue],
        is_sorted: bool = True,
    ) -> list[ModelEnumValue]:
        """Получение списка значений модели-перечисления моделей RegionalDataMartModelEnum.

        Args:
            entity_enums: Список значений модели-перечисления сущностей RegionalDataMartEntityEnum;
            is_sorted: Необходимость сортировки значений модели-перечисления RegionalDataMartModelEnum по полю
                order_number.
        """
        model_enums = set()

        for entity_enum_value in entity_enums:
            model_enums.update(cls.get_model_enums_from_value(value=entity_enum_value))

        model_enums = list(model_enums)

        if is_sorted:
            model_enums = sorted(model_enums, key=lambda value: value.order_number)

        return model_enums

    @classmethod
    def get_choices(cls) -> list[tuple[str, str]]:
        """Возвращает список кортежей из ключей и ключей перечисления сущностей."""
        return [(key, key) for key in sorted(cls.get_model_enum_keys())]

    @classmethod
    def extend(
        cls,
        key,
        title: str = '',
        entity: Type[BaseEntity] = None,
        main_model_enum: tuple = (),
        additional_model_enums: tuple = (),
        order_number: Optional[int] = None,
        *args,
        **kwargs,
    ):
        """Метод расширения модели-перечисления, например из плагина.

        Необходимо, чтобы сама модель-перечисление была расширяемой. Для этого необходимо, чтобы был установлен
        extensible = True в Meta.

        Args:
            key: ключ элемента перечисления, указывается заглавными буквами с разделителем нижнее подчеркивание
            title: название элемента перечисления
            entity: сущность, регистрируемая в модели-перечислении
            main_model_enum: основная модель РВД из модели-перечисления
            additional_model_enums: дополнительные модели РВД из модели-перечисления
            order_number: порядковый номер значения модели перечисления используемый при сортировке
            args: порядковые аргументы для создания значения модели перечисления
            kwargs: именованные аргументы для создания значения модели перечисления
        """
        try:
            order_number = cls._calculate_order_number(order_number=order_number)
        except ValueError as e:
            raise ValueError(f'Trying register entity "{entity.__name__}". {e.args[0]}')

        model_enum_value = ModelEnumValue(
            key=key,
            title=title,
            entity=entity,
            main_model_enum=main_model_enum,
            additional_model_enums=additional_model_enums,
            order_number=order_number,
            **kwargs,
        )

        setattr(cls, key, model_enum_value)



from functools import (
    partial,
)

from django.db.models import (
    F,
    Func,
    IntegerField,
    OuterRef,
    Q,
    Subquery,
)
from django.db.transaction import (
    atomic,
)
from m3.actions.exceptions import (
    ApplicationLogicException,
)
from m3.actions.results import (
    OperationResult,
)
from objectpack.actions import (
    BaseAction,
)

from educommon.async_task.actions import (
    RevokeAsyncTaskAction,
)
from educommon.async_task.models import (
    AsyncTaskStatus,
)
from educommon.utils.ui import (
    ChoicesFilter,
    DatetimeFilterCreator,
)

from edu_rdm_integration.core.enums import (
    CommandType,
)
from edu_rdm_integration.core.helpers import (
    make_download_link,
)
from edu_rdm_integration.core.registry.actions import (
    BaseCommandProgressPack,
    BaseStartTaskAction,
)
from edu_rdm_integration.rdm_entities.models import (
    RDMEntityEnum,
)
from edu_rdm_integration.stages.export_data.generators import (
    BaseFirstExportEntitiesDataCommandsGenerator,
)
from edu_rdm_integration.stages.export_data.models import (
    RDMExportingDataCommandProgress,
    RDMExportingDataStageStatus,
    RDMExportingDataSubStage,
    RDMExportingDataSubStageStatus,
)
from edu_rdm_integration.stages.export_data.registry.ui import (
    CreateExportCommandWindow,
    DetailExportCommandWindow,
    ExportCommandProgressListWindow,
)


class BaseExportingDataProgressPack(BaseCommandProgressPack):
    """Базовый пак команд экспорта данных сущностей РВД."""

    model = RDMExportingDataCommandProgress
    title = 'Экспорт данных сущностей РВД'

    add_window = CreateExportCommandWindow
    edit_window = DetailExportCommandWindow
    list_window = ExportCommandProgressListWindow

    need_check_permission = True

    select_related = ['task', 'task__status']

    list_sort_order = ('-created', 'entity__order_number', 'generation_id')
    date_filter = partial(DatetimeFilterCreator, model)

    columns = [
        {
            'data_index': 'entity.pk',
            'header': 'Сущность',
            'sortable': True,
            'filter': ChoicesFilter(
                choices=partial(RDMEntityEnum.get_choices),
                parser=str,
                lookup=lambda key: Q(entity=key) if key else Q(),
            ),
        },
        {
            'data_index': 'task.status.title',
            'header': 'Статус асинхронной задачи',
            'sortable': True,
            'filter': ChoicesFilter(
                choices=partial(AsyncTaskStatus.get_choices),
                parser=str,
                lookup='task__status_id',
            ),
        },
        {
            'data_index': 'type',
            'header': 'Тип команды',
            'filter': ChoicesFilter(
                choices=CommandType.get_choices(),
                parser=int,
                lookup='type',
            ),
            'width': 60,
        },
        {
            'data_index': 'finished_sub_stages',
            'header': 'Подэтапов <br> выполнено',
            'width': 50,
        },
        {
            'data_index': 'ready_sub_stages',
            'header': 'Подэтапов <br> подготовлено <br> к выгрузке',
            'width': 50,
        },
        {
            'data_index': 'process_errors_sub_stages',
            'header': 'Подэтапов <br> с ошибкой <br> обработки <br> запроса',
            'width': 50,
        },
        {
            'data_index': 'stage.status.key',
            'header': 'Статус экспорта',
            'sortable': True,
            'filter': ChoicesFilter(
                choices=partial(RDMExportingDataStageStatus.get_choices),
                parser=str,
                lookup=lambda key: Q(stage__status=key) if key else Q(),
            ),
            'width': 50,
        },
        {
            'data_index': 'stage.started_at',
            'header': 'Время начала экспорта',
            'sortable': True,
            'filter': date_filter('stage__started_at').filter,
        },
        {
            'data_index': 'log_url',
            'header': 'Ссылка на логи',
            'width': 60,
        },
        {
            'data_index': 'period_started_at',
            'header': 'Начало периода',
            'sortable': True,
            'filter': date_filter('period_started_at').filter,
        },
        {
            'data_index': 'period_ended_at',
            'header': 'Конец периода',
            'sortable': True,
            'filter': date_filter('period_ended_at').filter,
        },
        {
            'data_index': 'generation_id',
            'header': 'ID генерации',
            'sortable': True,
        },
        {
            'data_index': 'created',
            'header': 'Дата создания',
            'sortable': True,
        },
    ]

    _start_task_action_cls: BaseStartTaskAction
    _revoke_task_action_cls: RevokeAsyncTaskAction

    def __init__(self):
        super().__init__()

        self.start_task_action = self._start_task_action_cls()
        self.revoke_task_action = self._revoke_task_action_cls()
        self.prepare_sub_stage_for_export_action = PrepareSubStageForExportAction()

        self.actions.extend((self.start_task_action, self.revoke_task_action, self.prepare_sub_stage_for_export_action))

    def get_list_window_params(self, params, request, context):
        """Получает параметры окна списка."""
        params = super().get_list_window_params(params, request, context)

        params['revoke_url'] = self.revoke_task_action.get_absolute_url()
        params['start_task_url'] = self.start_task_action.get_absolute_url()
        params['sub_stage_for_export_url'] = self.prepare_sub_stage_for_export_action.get_absolute_url()

        return params

    def declare_context(self, action):
        """Декларирует контекст экшна."""
        context = super().declare_context(action)

        if action is self.save_action:
            context['period_started_at'] = {'type': 'datetime'}
            context['period_ended_at'] = {'type': 'datetime'}
            context['batch_size'] = {'type': 'int'}
        elif action in (self.start_task_action, self.prepare_sub_stage_for_export_action):
            context['commands'] = {'type': 'int_list'}
            context['queue_level'] = {'type': int, 'default': None}
        elif action is self.revoke_task_action:
            context['async_task_ids'] = {'type': 'str', 'default': ''}

        return context

    def get_rows_query(self, request, context):
        """Возвращает выборку из БД для получения списка данных."""
        query = super().get_rows_query(request, context)

        return query.annotate(
            finished_sub_stages=Subquery(
                RDMExportingDataSubStage.objects.filter(
                    stage_id=OuterRef('stage_id'),
                    status=RDMExportingDataSubStageStatus.FINISHED.key,
                )
                .annotate(
                    finished_count=Func(F('id'), function='Count'),
                )
                .values('finished_count'),
            ),
            ready_sub_stages=Subquery(
                RDMExportingDataSubStage.objects.filter(
                    stage_id=OuterRef('stage_id'),
                    status=RDMExportingDataSubStageStatus.READY_FOR_EXPORT.key,
                )
                .annotate(
                    ready_count=Func(F('id'), function='Count', output_field=IntegerField()),
                )
                .values('ready_count'),
            ),
            process_errors_sub_stages=Subquery(
                RDMExportingDataSubStage.objects.filter(
                    stage_id=OuterRef('stage_id'),
                    status=RDMExportingDataSubStageStatus.PROCESS_ERROR.key,
                )
                .annotate(process_errors_count=Func(F('id'), function='Count', output_field=IntegerField()))
                .values('process_errors_count'),
            ),
        )

    def prepare_row(self, obj, request, context):
        """Подготовка данных для отображения в реестре."""
        obj.log_url = make_download_link(obj.logs_link)

        return obj

    @atomic
    def save_row(self, obj, create_new, request, context, *args, **kwargs):
        """Сохраняет объекты."""
        commands = BaseFirstExportEntitiesDataCommandsGenerator(
            entities=[obj.entity_id],
            period_started_at=context.period_started_at,
            period_ended_at=context.period_ended_at,
            batch_size=context.batch_size,
        ).generate()

        if not commands:
            raise ApplicationLogicException(
                f'Недостаточно данных для выгрузки по модели "{obj.entity_id}" '
                f'в указанный период с {context.period_started_at} по {context.period_ended_at}. '
                f'Проверьте наличие данных в указанном временном диапазоне.'
            )

        for command in commands:
            obj = self.model(
                entity_id=obj.entity_id,
                period_started_at=command['period_started_at'],
                period_ended_at=command['period_ended_at'],
                generation_id=command['generation_id'],
                type=CommandType.MANUAL,
            )
            super().save_row(obj, create_new, request, context, *args, **kwargs)


class PrepareSubStageForExportAction(BaseAction):
    """Смена статусов у подэтапов для переотправки."""

    def run(self, request, context):
        """Обновление статусов подэтапов не принятых витриной."""
        command_ids = context.commands
        stage_ids = RDMExportingDataCommandProgress.objects.filter(id__in=command_ids).values_list(
            'stage_id', flat=True
        )

        updated_count = RDMExportingDataSubStage.objects.filter(
            stage_id__in=stage_ids,
            status=RDMExportingDataSubStageStatus.PROCESS_ERROR.key,
        ).update(status=RDMExportingDataSubStageStatus.READY_FOR_EXPORT.key)
        if updated_count:
            message = f'Будет переотправлено {updated_count} подэтапов.'
        else:
            message = 'Подэтапов для переотправления не найдено.'

        return OperationResult(success=True, message=message)



import logging
from abc import (
    ABC,
    abstractmethod,
)
from datetime import (
    date,
)
from typing import (
    Optional,
    Union,
)

from django.db.models import (
    DateField,
    Expression,
    F,
)
from django.db.models.functions import (
    Cast,
)

from educommon.integration_entities.enums import (
    EntityLogOperation,
)

from edu_rdm_integration.core.utils import (
    make_passed_datetime_from_today,
)


# Кортеж операций для обновления данных
UPDATED_OPERATIONS = (EntityLogOperation.CREATE, EntityLogOperation.UPDATE)
ALL_OPERATIONS = UPDATED_OPERATIONS + (EntityLogOperation.DELETE,)
TODAY_EXPR: Union[date, Expression, F] = Cast(make_passed_datetime_from_today(), output_field=DateField())
THREE_YEARS_AGO_DATE_EXPR: Union[date, Expression, F] = Cast(
    make_passed_datetime_from_today(years=3), output_field=DateField()
)


class BaseOperationData(ABC):
    """Базовый класс операций с данными."""

    def __init__(self, **kwargs):
        # Идентификатор команды для передачи сигналу manager_created
        self.command_id: Optional[int] = kwargs.get('command_id')

        self._file_handler: Optional[logging.FileHandler] = None

        self._add_file_handler()

    @property
    @abstractmethod
    def _log_file_path(self) -> str:
        """Путь до лог файла."""

    def _add_file_handler(self) -> None:
        """Добавляет обработчик логов."""
        if self.command_id:
            self._file_handler = logging.FileHandler(self._log_file_path)

            logging.getLogger('info_logger').addHandler(self._file_handler)
            logging.getLogger('exception_logger').addHandler(self._file_handler)

    def _remove_file_handler(self) -> None:
        """Удаляет обработчик логов."""
        if self._file_handler:
            logging.getLogger('info_logger').removeHandler(self._file_handler)
            logging.getLogger('exception_logger').removeHandler(self._file_handler)

            self._file_handler.close()



from edu_function_tools.results import (
    EduFunctionResult,
    EduRunnerResult,
)


class BaseExportDataRunnerResult(EduRunnerResult):
    """Базовый класс результата работы ранера функций выгрузки данных для интеграции с РВД."""


class BaseExportDataFunctionResult(EduFunctionResult):
    """Базовый класс результата функции выгрузки данных для интеграции с РВД."""



import os
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Iterable,
    Union,
)

from django.conf import (
    settings,
)
from django.db.models import (
    CharField,
    OuterRef,
    Subquery,
    Value,
)
from django.db.models.functions import (
    Cast,
    Coalesce,
    Least,
)
from django.db.transaction import (
    atomic,
)

from educommon import (
    logger,
)

from edu_rdm_integration.apps import (
    EduRDMIntegrationConfig,
)


if TYPE_CHECKING:
    from typing import (
        NamedTuple,
    )

    from django.db.models import (
        Model,
    )

    from edu_rdm_integration.rdm_models.models import (
        BaseRDMModel,
    )
    from edu_rdm_integration.stages.export_data.functions.base.managers import (
        BaseExportDataRunnerManager,
    )


@atomic
def set_failed_status_suspended_exporting_data_stages() -> dict[str, int]:
    """Установить статус 'Завершено с ошибками' для зависших этапов и подэтапов экспорта.

    Экспорт считается зависшим в случае если за определенное в параметре RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT время,
    отсутствуют изменения в связанных подэтапах. Параметр RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT определяется
    в настройках приложения.
    """
    from edu_rdm_integration.stages.export_data.models import (
        RDMExportingDataStage,
        RDMExportingDataStageStatus,
        RDMExportingDataSubStage,
        RDMExportingDataSubStageStatus,
    )

    changed_status_result = {
        'change_stage_count': 0,
        'change_sub_stage_count': 0,
    }

    current_datetime = datetime.now()
    suspended_time_at = current_datetime - timedelta(minutes=settings.RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT)

    suspended_stage_ids = set(
        RDMExportingDataStage.objects.annotate(
            last_sub_stage_started_at=Coalesce(
                Subquery(
                    RDMExportingDataSubStage.objects.filter(stage_id=OuterRef('pk'))
                    .values('started_at')
                    .order_by('-started_at')[:1]
                ),
                Value(datetime.combine(date.min, time.min)),
            )
        )
        .filter(
            last_sub_stage_started_at__lt=suspended_time_at,
            status__in=(
                RDMExportingDataStageStatus.CREATED.key,
                RDMExportingDataStageStatus.IN_PROGRESS.key,
            ),
        )
        .values_list('pk', flat=True)
    )

    if suspended_stage_ids:
        logger.info(f'find suspended ExportingDataStage: {", ".join(map(str, suspended_stage_ids))}..')

        change_stage_count = RDMExportingDataStage.objects.filter(
            pk__in=suspended_stage_ids,
        ).update(
            status=RDMExportingDataStageStatus.FAILED.key,
            ended_at=current_datetime,
        )

        change_sub_stage_count = RDMExportingDataSubStage.objects.filter(
            stage_id__in=suspended_stage_ids,
        ).update(
            status=RDMExportingDataSubStageStatus.FAILED.key,
            ended_at=current_datetime,
        )

        changed_status_result.update(
            {
                'change_stage_count': change_stage_count,
                'change_sub_stage_count': change_sub_stage_count,
            }
        )

    return changed_status_result


def get_exporting_managers_max_period_ended_dates(
    exporting_managers: Iterable['BaseExportDataRunnerManager'],
) -> dict[str, 'datetime']:
    """Возвращает дату и время последнего успешного этапа экспорта для менеджеров Функций экспорта."""
    from edu_rdm_integration.stages.export_data.models import (
        RDMExportingDataStage,
        RDMExportingDataStageStatus,
    )

    managers_last_period_ended = (
        RDMExportingDataStage.objects.filter(
            manager_id__in=[manager.uuid for manager in exporting_managers],
            id=Subquery(
                RDMExportingDataStage.objects.filter(
                    manager_id=OuterRef('manager_id'),
                    status_id=RDMExportingDataStageStatus.FINISHED.key,
                )
                .order_by('-id')
                .values('id')[:1]
            ),
        )
        .annotate(
            str_manager_id=Cast('manager_id', output_field=CharField()),
            last_period_ended_at=Least('period_ended_at', 'started_at'),
        )
        .values_list(
            'str_manager_id',
            'last_period_ended_at',
        )
    )

    return {manager_id: last_period_ended_at for manager_id, last_period_ended_at in managers_last_period_ended}


def get_exporting_data_stage_attachment_path(instance, filename):
    """Возвращает путь до файла-вложения в этап выгрузки данных сущности.

    Args:
        instance: объект ExportingDataStage
        filename: имя загружаемого файла

    Returns:
        Строковое представление пути
    """
    datetime_now = datetime.now()

    return os.path.join(
        settings.UPLOADS,
        EduRDMIntegrationConfig.label,
        datetime_now.strftime('%Y/%m/%d'),
        instance.exporting_data_sub_stage.__class__.__name__.lower(),
        str(instance.exporting_data_sub_stage_id),
        str(instance.operation),
        filename,
    )


def set_entity_field_by_model_object(
    entity: 'BaseRDMModel', model_object: Union['Model', 'NamedTuple'], mapping: dict[str, str]
) -> None:
    """Обновление значений полей сущности по измененным полям модели.

    Args:
        entity: Выгружаемая сущность;
        model_object: Объект модели с измененными полями;
        mapping: Словарь маппинга полей модели к полям сущности
    """
    for model_field, entity_field in mapping.items():
        if hasattr(model_object, model_field):
            setattr(entity, entity_field, getattr(model_object, model_field))


def get_isoformat_timezone():
    """Возвращает временную зонну в ISO представлении."""
    return datetime.now().astimezone().isoformat()[-6:]



from django.db.models import (
    CASCADE,
    PROTECT,
    SET_NULL,
    BooleanField,
    CharField,
    DateTimeField,
    ForeignKey,
    SmallIntegerField,
)
from m3.db import (
    BaseObjectModel,
)
from uploader_client.models import (
    Entry,
)

from educommon.django.db.mixins import (
    ReprStrPreModelMixin,
)
from m3_db_utils.models import (
    ModelEnumValue,
    TitledIntegerModelEnum,
)

from edu_rdm_integration.stages.export_data.models import (
    RDMExportingDataSubStage,
    RDMExportingDataSubStageAttachment,
)
from edu_rdm_integration.stages.upload_data.enums import (
    FileUploadStatusEnum,
)


class RDMRequestStatus(TitledIntegerModelEnum):
    """Модель-перечисление статусов загрузки данных в Витрину."""

    UPLOAD_TO_BUFFER = ModelEnumValue(
        value=-1,
        title='Загрузка данных в буффер',
    )

    BUFFERED = ModelEnumValue(
        value=0,
        title='Запрос буфферизирован',
    )

    WAIT_FOR_OPEN_DELTA = ModelEnumValue(
        value=1,
        title='Ожидает открытия дельты',
    )

    IN_PROCESSING = ModelEnumValue(
        value=2,
        title='В обработке',
    )

    SUCCESSFULLY_PROCESSED = ModelEnumValue(
        value=3,
        title='Успешно обработан',
    )

    FAILED_PROCESSING = ModelEnumValue(
        value=4,
        title='Ошибка обработки запроса',
    )

    REQUEST_ID_NOT_FOUND = ModelEnumValue(
        value=5,
        title='Идентификатор запроса не обнаружен',
    )

    FORMAT_LOGICAL_CONTROL = ModelEnumValue(
        value=6,
        title='Форматно-логический контроль',
    )

    FLC_ERROR = ModelEnumValue(
        value=7,
        title='Ошибки ФЛК',
    )

    class Meta:
        db_table = 'rdm_request_status'
        verbose_name = 'Статус загрузки данных в Витрину'
        verbose_name_plural = 'Статусы загрузки данных в Витрину'


class RDMExportingDataSubStageUploaderClientLog(ReprStrPreModelMixin, BaseObjectModel):
    """Связь лога Загрузчика данных с подэтапом выгрузки данных."""

    entry = ForeignKey(
        to=Entry,
        verbose_name='Лог запроса и ответа',
        on_delete=SET_NULL,
        null=True,
        related_name='uploader_client_log',
    )

    sub_stage = ForeignKey(
        to=RDMExportingDataSubStage,
        verbose_name='Подэтап выгрузки данных',
        on_delete=CASCADE,
    )

    attachment = ForeignKey(
        to=RDMExportingDataSubStageAttachment,
        verbose_name='Прикрепленный файл',
        on_delete=CASCADE,
    )

    request_id = CharField(
        verbose_name='Id запроса загрузки в витрину',
        max_length=100,
        blank=True,
        db_index=True,
    )

    is_emulation = BooleanField(
        verbose_name='Включен режим эмуляции',
        default=False,
    )

    file_upload_status = SmallIntegerField(
        verbose_name='Общий статус загрузки файла в витрину',
        choices=FileUploadStatusEnum.get_choices(),
        null=True,
        blank=True,
    )

    created = DateTimeField(
        verbose_name='Дата создания',
        auto_now_add=True,
        null=True,
        blank=True,
        db_index=True,
    )
    modified = DateTimeField(
        verbose_name='Дата изменения',
        auto_now=True,
        null=True,
        blank=True,
        db_index=True,
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage_uploader_client_log'
        verbose_name = 'Лог запроса подэтапа выгрузки данных'
        verbose_name_plural = 'Лог запроса подэтапа выгрузки данных'


class RDMUploadStatusRequestLog(ReprStrPreModelMixin, BaseObjectModel):
    """Модель связывающая статусы, загрузку файла в витрину и http-запрос к витрине."""

    upload = ForeignKey(
        verbose_name='Cвязь запроса статуса с загрузкой файла в витрину',
        to=RDMExportingDataSubStageUploaderClientLog,
        on_delete=CASCADE,
    )
    entry = ForeignKey(
        verbose_name='Cвязь запроса статуса с запросом в витрину',
        to=Entry,
        null=True,
        on_delete=SET_NULL,
        related_name='upload_status_request_log',
    )
    request_status = ForeignKey(
        verbose_name='Статус запроса в витрине',
        to=RDMRequestStatus,
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    class Meta:
        db_table = 'rdm_upload_status_request_log'
        verbose_name = 'Лог запроса подэтапа выгрузки данных'
        verbose_name_plural = 'Логи запроса подэтапа выгрузки данных'



from educommon.async_task.models import (
    AsyncTaskType,
)

from edu_rdm_integration.core.consts import (
    TASK_QUEUE_NAME,
)
from edu_rdm_integration.core.helpers import (
    save_command_log_link,
)
from edu_rdm_integration.stages.export_data.models import (
    RDMExportingDataCommandProgress,
)
from edu_rdm_integration.stages.export_data.operations import (
    ExportEntitiesData,
)


class ExportCommandMixin:
    """Класс-примесь для запуска команды выгрузки сущностей."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Экспорт данных сущностей РВД'
    task_type = AsyncTaskType.SYSTEM

    def get_export_command(self, command_id: int) -> RDMExportingDataCommandProgress:
        """Возвращает экземпляр модели команды запуска."""
        command = RDMExportingDataCommandProgress.objects.get(id=command_id)

        return command

    def run_export_command(self, command: RDMExportingDataCommandProgress) -> None:
        """Запуск команды выгрузки."""
        ExportEntitiesData(
            entities=(command.entity_id,),
            period_started_at=command.period_started_at,
            period_ended_at=command.period_ended_at,
            command_id=command.id,
        ).export()

    def save_export_command_logs(self, command_id: int, log_dir: str):
        """Сохранение ссылки на файл логов в команде."""
        try:
            command = self.get_export_command(command_id)
        except RDMExportingDataCommandProgress.DoesNotExist:
            command = None

        if command:
            save_command_log_link(command, log_dir)



from abc import (
    ABCMeta,
)

from edu_function_tools.presenters import (
    EduResultPresenter,
)


class BaseCollectingExportedDataResultPresenter(EduResultPresenter, metaclass=ABCMeta):
    """Презентер результата работы функций сбора данных для интеграции с "Региональная витрина данных"."""



from educommon.async_task.models import (
    AsyncTaskType,
)

from edu_rdm_integration.core.consts import (
    TASK_QUEUE_NAME,
)
from edu_rdm_integration.core.helpers import (
    save_command_log_link,
)
from edu_rdm_integration.stages.collect_data.models import (
    RDMCollectingDataCommandProgress,
)
from edu_rdm_integration.stages.collect_data.operations import (
    BaseCollectModelsDataByGeneratingLogs,
)


class CollectCommandMixin:
    """Класс-примесь для запуска команды сборки моделей."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Сбор данных моделей РВД'
    task_type = AsyncTaskType.SYSTEM

    def get_collect_command(self, command_id: int) -> RDMCollectingDataCommandProgress:
        """Возвращает экземпляр модели команды запуска."""
        command = RDMCollectingDataCommandProgress.objects.get(id=command_id)

        return command

    def get_collect_models_class(self):
        """Возвращает класс для сбора данных."""
        return BaseCollectModelsDataByGeneratingLogs

    def run_collect_command(self, command) -> None:
        """Запуск команды сбора."""
        collect_models_data_class = self.get_collect_models_class()
        collect_models_data_class(
            models=(command.model_id,),
            logs_period_started_at=command.logs_period_started_at,
            logs_period_ended_at=command.logs_period_ended_at,
            command_id=command.id,
            institute_ids=tuple(command.institute_ids or ()),
        ).collect()

    def save_collect_command_logs(self, command_id: int, log_dir: str):
        """Сохранение ссылки на файл логов в команде."""
        try:
            command = self.get_collect_command(command_id)
        except RDMCollectingDataCommandProgress.DoesNotExist:
            command = None

        if command:
            save_command_log_link(command, log_dir)



# Generated by Django 3.2.24 on 2025-06-28 16:04

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('edu_rdm_integration_entities', '0001_initial'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='TransferredEntity',
                    fields=[
                        (
                            'id',
                            models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
                        ),
                        (
                            'export_enabled',
                            models.BooleanField(default=True, verbose_name='Включение экспорта для сущности'),
                        ),
                        (
                            'queue_level',
                            models.PositiveIntegerField(
                                choices=[(1, 'Быстрая'), (2, 'Основная'), (3, 'Долгая')],
                                default=2,
                                verbose_name='Уровень очереди',
                            ),
                        ),
                        (
                            'entity',
                            models.OneToOneField(
                                on_delete=django.db.models.deletion.CASCADE,
                                to='edu_rdm_integration_entities.regionaldatamartentityenum',
                                verbose_name='Сущность',
                            ),
                        ),
                    ],
                    options={
                        'verbose_name': 'Сущность, по которой должен быть произведен сбор и экспорт данных',
                        'verbose_name_plural': 'Сущности, по которым должен быть произведен сбор и экспорт данных',
                        'db_table': 'rdm_transferred_entity',
                    },
                ),
            ],
            database_operations=[],
        ),
    ]



from django.apps import (
    AppConfig,
)


class RDMExportDataRegistryAppConfig(AppConfig):
    """Приложение для организации экспорта данных моделей РВД из интерфейса."""

    name = 'edu_rdm_integration.stages.export_data.registry'
    label = 'edu_rdm_integration_export_data_registry'
    verbose_name = 'Экспорт данных моделей РВД из интерфейса'



from typing import (
    Optional,
)

from django.core.exceptions import (
    FieldDoesNotExist,
)
from django.db.models import (
    Subquery,
)

from edu_rdm_integration.rdm_models.models import (
    RDMModelEnum,
)
from edu_rdm_integration.stages.collect_data.models import (
    RDMCollectingDataCommandProgress,
    RDMCollectingDataStage,
    RDMCollectingDataSubStage,
)
from educommon.utils.seqtools import (
    make_chunks,
)

from web_edu.plugins.regional_data_mart_integration.models import (
    LessonClass,
)

from .base import (
    BaseServiceOutdatedDataCleaner,
)
from .consts import (
    OLD_RDM_MODEL,
    UNION_CHUNK_SIZE,
)


class CollectingDataSubStageCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка подэтапов сбора данных, которые не ссылаются ни на одну модель РВД.

    Подход:
        - Проходим по всем моделям, зарегистрированным в RDMModelEnum, и собираем
            значения полей collecting_sub_stage_id (если модель содержит такое поле).
        - Объединяем запросы по моделям в UNION, получая набор валидных id.
        - Удаляем те подэтапы, id которых отсутствуют в полученном наборе.
    """

    model = RDMCollectingDataSubStage

    def _get_valid_substage_ids_subquery(self) -> Optional[Subquery]:
        """Подзапрос, возвращающий все допустимые collecting_sub_stage_id из моделей, описанных в RDMModelEnum."""
        model_enum_values = RDMModelEnum.get_model_enum_values()
        all_model = [model_enum.model for model_enum in model_enum_values] + OLD_RDM_MODEL
        chunk_queries = []

        for enum_values_chunk in make_chunks(all_model, UNION_CHUNK_SIZE, is_list=True):
            qs_list = []
            for model_cls in enum_values_chunk:
                if model_cls in [LessonClass]:
                    continue
                try:
                    model_cls._meta.get_field('collecting_sub_stage_id')
                except FieldDoesNotExist:
                    continue

                qs_list.append(model_cls.objects.values('collecting_sub_stage_id'))

            if qs_list:
                chunk_union = qs_list[0].union(*qs_list[1:])
                chunk_queries.append(chunk_union)

        if not chunk_queries:
            return

        # Объединяем все чанки в один общий UNION
        full_union = chunk_queries[0].union(*chunk_queries[1:])

        return Subquery(full_union)

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных.

         Удаляем подэтапы, которых нет в объединённом наборе валидных collecting_sub_stage_id
        (т.е. подэтапы, не используемые ни одной моделью данных).
        """
        conditions = ''
        subquery = self._get_valid_substage_ids_subquery()
        if subquery:
            conditions = f"""
                NOT EXISTS (
                    SELECT collecting_sub_stage_id
                    FROM ({str(subquery.query)}) AS valid
                    WHERE valid.collecting_sub_stage_id = tbl.id
                )
            """

        return conditions


class CollectingDataStageCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка этапов сбора данных, у которых нет связанных подэтапов."""

    model = RDMCollectingDataStage

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        sub_stage_table = CollectingDataSubStageCleaner.get_table_name()

        return self.get_orphan_reference_condition(sub_stage_table, 'stage_id')


class CollectingDataCommandProgressCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка устаревших хранящихся задач по сбору данных."""

    model = RDMCollectingDataCommandProgress

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        stage_table = CollectingDataStageCleaner.get_table_name()
        conditions = [
            'stage_id IS NULL',
            f'({self.get_status_condition(stage_table, "id", "FINISHED", 7, "stage_id")})',
            f'({self.get_status_condition(stage_table, "id", "FAILED", 30, "stage_id")})',
            f'({self.get_orphan_reference_condition(stage_table, "id", "stage_id")})',
        ]

        return " OR ".join(conditions)



# Generated by Django 3.2.24 on 2025-07-04 07:25

from django.db import (
    migrations,
)


class Migration(migrations.Migration):
    dependencies = [
        ('rdm_collect_and_export_data', '0002_auto_20250204_1413'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name='edurdmexportdatacommandprogress',
                    name='entity',
                ),
                migrations.RemoveField(
                    model_name='edurdmexportdatacommandprogress',
                    name='stage',
                ),
                migrations.RemoveField(
                    model_name='edurdmexportdatacommandprogress',
                    name='task',
                ),
                migrations.DeleteModel(
                    name='EduRdmCollectDataCommandProgress',
                ),
                migrations.DeleteModel(
                    name='EduRdmExportDataCommandProgress',
                ),
            ],
            database_operations=[],
        ),
    ]



from django.conf import (
    settings,
)
from m3 import (
    OperationResult,
)
from m3.actions import (
    ControllerCache,
)
from m3_ext.ui.containers import (
    ExtGridCheckBoxSelModel,
)
from objectpack.actions import (
    BaseAction,
    ObjectPack,
    SelectorWindowAction,
)

from edu_rdm_integration.pipelines.transfer.models import (
    TransferredEntity,
)
from edu_rdm_integration.pipelines.transfer.ui import (
    TransferredEntityEditWindow,
    TransferredEntityListWindow,
)
from edu_rdm_integration.rdm_entities.models import (
    RDMEntityEnum,
)


class EntitySelectPack(ObjectPack):
    """Пак выбора сущностей для сбора и экспорта данных."""

    title = 'Сущность РВД'
    model = RDMEntityEnum
    _is_primary_for_model = False

    list_sort_order = ('order_number',)

    columns = [
        {
            'data_index': 'key',
            'header': 'Сущность',
        },
        {
            'data_index': 'title',
            'header': 'Описание',
        },
    ]

    def get_rows_query(self, request, context):
        """Возвращает выборку из БД для получения списка данных.

        Ранее выбранные сущности не отображаются в списке.
        """
        query = super().get_rows_query(request, context)

        return query.filter(transferredentity__isnull=True)


class TransferredEntityAddWindowAction(SelectorWindowAction):
    """Экшн показа окна выбора сущностей для сбора и экспорта данных."""

    def configure_action(self, request, context):
        """Конфигурирует экшн."""
        self.data_pack = ControllerCache.find_pack(EntitySelectPack)
        self.callback_url = self.parent.save_action.absolute_url()

    def configure_window(self, win, request, context):
        """Конфигурирует окно выбора."""
        win.grid.sm = ExtGridCheckBoxSelModel()
        win.grid.store.id_property = 'key'

        return win


class TransferredEntityExportChangeAction(BaseAction):
    """Экшен смены статуса экспорта выбранных сущностей."""

    def context_declaration(self):
        """Объявляет контекст экшна."""
        return {'ids': {'type': 'int_list', 'default': []}, 'export_enabled': {'type': 'boolean', 'default': True}}

    def run(self, request, context):
        """Обеспечивает выполнение запроса."""
        self.parent.model.objects.filter(id__in=context.ids).update(export_enabled=context.export_enabled)

        return OperationResult(success=True)


class TransferredEntitySaveAction(BaseAction):
    """Экшн сохранения выбранных сущностей для сбора и экспорта данных."""

    def context_declaration(self):
        """Объявляет контекст экшна."""
        return {'id': {'type': 'str_list', 'default': []}, self.parent.id_param_name: {'type': 'int'}}

    def run(self, request, context):
        """Обеспечивает выполнение запроса."""
        obj_id = getattr(context, self.parent.id_param_name)
        if obj_id:
            queue_level = getattr(context, 'queue_level', None)
            if queue_level:
                self.parent.model.objects.filter(id=obj_id).exclude(queue_level=queue_level).update(
                    queue_level=queue_level
                )
        else:
            self.parent.model.objects.bulk_create([self.parent.model(entity_id=key) for key in context.id])

        return OperationResult(success=True)


class TransferredEntityPack(ObjectPack):
    """Пак сущностей, по которым должен быть произведен сбор и экспорт данных."""

    title = 'Сущности для сбора и экспорта данных'
    model = TransferredEntity

    can_delete = True

    list_sort_order = ('entity__order_number',)

    need_check_permission = True
    list_window = TransferredEntityListWindow
    edit_window = TransferredEntityEditWindow

    columns = [
        {
            'data_index': 'entity.key',
            'header': 'Сущность',
        },
        {
            'data_index': 'entity.title',
            'header': 'Описание',
        },
        {
            'data_index': 'no_export',
            'header': 'Отключение экспорта',
        },
        {
            'data_index': 'queue_level',
            'header': 'Уровень очереди',
        },
    ]

    def __init__(self):
        super().__init__()

        self.add_window_action = TransferredEntityAddWindowAction()
        self.replace_action('new_window_action', self.add_window_action)

        self.save_entity_action = TransferredEntitySaveAction()
        self.replace_action('save_action', self.save_entity_action)

        self.export_change_action = TransferredEntityExportChangeAction()

        self.actions.append(self.export_change_action)

    def configure_grid(self, grid, *args, **kwargs):
        """Конфигурирует грид."""
        super().configure_grid(grid, *args, **kwargs)

        grid.sm = ExtGridCheckBoxSelModel()

    def extend_menu(self, menu):
        """Расширяет главное меню."""
        if settings.RDM_MENU_ITEM:
            return menu.SubMenu(
                'Администрирование',
                menu.SubMenu(
                    'Региональная витрина данных',
                    menu.Item(
                        self.title,
                        self.list_window_action,
                    ),
                    icon='menu-dicts-16',
                ),
            )



from django.apps import (
    apps,
)
from django.db import (
    models,
)

from edu_rdm_integration.core.utils import (
    register_classes,
)
from edu_rdm_integration.rdm_models.mixins import (
    ModelEnumRegisterMixin,
)


def is_register_model(model: models.Model) -> bool:
    """Проверяет, является ли класс регистрируемой моделью."""
    return issubclass(model, ModelEnumRegisterMixin)


def register_models() -> None:
    """Регистрирует модели в RegionalDataMartModelEnum."""
    register_classes([m for m in apps.get_models() if is_register_model(m)])



from abc import (
    ABCMeta,
)

from edu_rdm_integration.stages.collect_data.functions.base.functions import (
    BaseCollectingCalculatedDataFunction,
)
from edu_rdm_integration.stages.collect_data.functions.non_calculated.base.helpers import (
    BaseCollectingExportedDataFunctionHelper,
)
from edu_rdm_integration.stages.collect_data.functions.non_calculated.base.results import (
    BaseCollectingExportedDataFunctionResult,
)
from edu_rdm_integration.stages.collect_data.functions.non_calculated.base.validators import (
    BaseCollectingExportedDataFunctionValidator,
)


class BaseCollectingExportedDataFunction(BaseCollectingCalculatedDataFunction, metaclass=ABCMeta):
    """Базовый класс функций сбора данных для интеграции с "Региональная витрина данных"."""

    def _prepare_helper_class(self) -> type[BaseCollectingExportedDataFunctionHelper]:
        """Возвращает класс помощника функции."""
        return BaseCollectingExportedDataFunctionHelper

    def _prepare_validator_class(self) -> type[BaseCollectingExportedDataFunctionValidator]:
        """Возвращает класс валидатора функции."""
        return BaseCollectingExportedDataFunctionValidator

    def _prepare_result_class(self) -> type[BaseCollectingExportedDataFunctionResult]:
        """Возвращает класс результата функции."""
        return BaseCollectingExportedDataFunctionResult



from abc import (
    ABCMeta,
    abstractmethod,
)


class AbstractCache(metaclass=ABCMeta):
    """Абстрактный интерфейс для кеша отправки."""

    @abstractmethod
    def get(self, key, default=None, **kwargs):
        """Возвращает значение из кеша по ключу."""

    @abstractmethod
    def set(self, key, value, timeout=None, **kwargs):
        """Сохраняет значение в кеш по ключу."""

    @abstractmethod
    def lock(self, name, timeout=None, **kwargs):
        """Захватывает блокировку."""



# Generated by Django 3.2.24 on 2025-10-08 14:43

from django.db import (
    migrations,
)

from edu_rdm_integration.stages.utils import (
    update_foreign_key_constraint,
)


def apply_fk_updates(apps, schema_editor):
    """
    Применяет обновления внешних ключей для конкретных моделей.

    Здесь задаются таблицы и поведение при удалении записей,
    чтобы синхронизировать фактическое состояние БД с логикой моделей.
    """
    RDMCollectingDataSubStage = apps.get_model('edu_rdm_integration_collect_data_stage', 'RDMCollectingDataSubStage')
    RDMCollectingDataStage = apps.get_model('edu_rdm_integration_collect_data_stage', 'RDMCollectingDataStage')

    # Модель: RDMCollectingDataSubStage
    # Поле previous_id должно иметь ON DELETE SET NULL
    update_foreign_key_constraint(
        table_name=RDMCollectingDataSubStage._meta.db_table,
        field_name="previous_id",
        target_table=RDMCollectingDataSubStage._meta.db_table,
        on_delete="SET NULL",
    )

    RDMCollectingDataCommandProgress = apps.get_model('edu_rdm_integration_collect_data_stage', 'RDMCollectingDataCommandProgress')
    # Модель: RDMCollectingDataCommandProgress
    # Поле stage_id должно иметь ON DELETE SET NULL
    update_foreign_key_constraint(
        table_name=RDMCollectingDataCommandProgress._meta.db_table,
        field_name="stage_id",
        target_table=RDMCollectingDataStage._meta.db_table,
        on_delete="SET NULL",
    )

    # Таблица: rdm_collecting_data_command_progress
    # Поле stage_id должно иметь ON DELETE SET NULL
    update_foreign_key_constraint(
        table_name='rdm_collecting_data_command_progress',
        field_name="stage_id",
        target_table=RDMCollectingDataStage._meta.db_table,
        on_delete="SET NULL",
    )


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration_collect_data_stage', '0005_alter_rdmcollectingdatasubstage_previous'),
    ]

    operations = [
        migrations.RunPython(apply_fk_updates, reverse_code=migrations.RunPython.noop),
    ]



# Generated by Django 3.2.24 on 2025-10-08 14:17

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration_collect_data_stage', '0004_auto_20250704_0825'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rdmcollectingdatasubstage',
            name='previous',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='edu_rdm_integration_collect_data_stage.rdmcollectingdatasubstage', verbose_name='Предыдущий сбор данных'),
        ),
    ]



from web_edu.plugins.regional_data_mart_integration.models.homework import Homework
from web_edu.plugins.regional_data_mart_integration.models.homework_material import HomeworkMaterial
from web_edu.plugins.regional_data_mart_integration.models.homework_student import HomeworkStudent


UNION_CHUNK_SIZE = 5

OLD_RDM_MODEL = [
    HomeworkStudent,
    HomeworkMaterial,
    Homework,
]


from typing import (
    Any,
    NamedTuple,
)

from django.db import (
    models,
)

from m3_db_utils.mixins import (
    BaseEnumRegisterMixin,
)
from m3_django_compatibility import (
    classproperty,
)

from edu_rdm_integration.core.typing import (
    MODEL_TYPE_VAR,
)
from edu_rdm_integration.core.utils import (
    camel_to_underscore,
)
from edu_rdm_integration.rdm_models.models import (
    RDMModelEnum,
)


class ModelEnumRegisterMixin(BaseEnumRegisterMixin):
    """Миксин, для регистрации модели в RegionalDataMartModelEnum."""

    enum = RDMModelEnum
    """Модель-перечисление в которую регистрируется модель."""

    creating_trigger_models: tuple[models.Model, ...] = ()
    """Перечень моделей по которым генерируются логи."""

    loggable_models: tuple[models.Model, ...] = ()
    """Перечень моделей по которым собираются логи."""

    @classproperty
    def key(self) -> str:
        """Ключ сущности в модели-перечислении RegionalDataMartEntityEnum."""
        return camel_to_underscore(self.__name__).upper()

    @classproperty
    def title(self):
        """Название сущности."""
        return self._meta.verbose_name

    @classmethod
    def get_register_params(cls) -> dict[str, Any]:
        """Параметры для регистрации сущности в модели перечислении RegionalDataMartEntityEnum."""
        register_params = super().get_register_params()
        register_params['model'] = cls
        register_params['creating_trigger_models'] = cls.creating_trigger_models
        register_params['loggable_models'] = cls.loggable_models

        return register_params


class FromNamedTupleMixin:
    """Миксин получения экземпляра модели из получаемого кэша значений."""

    @classmethod
    def from_namedtuple(cls: type[MODEL_TYPE_VAR], namedtuple: NamedTuple) -> MODEL_TYPE_VAR:
        """Создает экземпляр класса из NamedTuple."""
        return cls(**{field: getattr(namedtuple, field) for field in [f.column for f in cls._meta.get_fields()]})



from django.apps import (
    AppConfig,
)


class RDMCoreRegistryAppConfig(AppConfig):
    """Приложение с базовым функционалом для создания реестров РВД."""

    name = 'edu_rdm_integration.core.registry'
    label = 'edu_rdm_integration_core_registry'
    verbose_name = 'Базовая часть реестров РВД'



from edu_rdm_integration.stages.collect_data.functions.base.runners import (
    BaseCollectingDataRunner,
)
from edu_rdm_integration.stages.collect_data.functions.non_calculated.base.helpers import (
    BaseCollectingExportedDataRunnerHelper,
)
from edu_rdm_integration.stages.collect_data.functions.non_calculated.base.results import (
    BaseCollectingExportedDataRunnerResult,
)
from edu_rdm_integration.stages.collect_data.functions.non_calculated.base.validators import (
    BaseCollectingExportedDataRunnerValidator,
)


class BaseCollectingExportedDataRunner(BaseCollectingDataRunner):
    """Базовый класс ранеров функций сбора данных для интеграции с "Региональная витрина данных"."""

    def _prepare_helper_class(self) -> type[BaseCollectingExportedDataRunnerHelper]:
        """Возвращает класс помощника ранера функции."""
        return BaseCollectingExportedDataRunnerHelper

    def _prepare_validator_class(self) -> type[BaseCollectingExportedDataRunnerValidator]:
        """Возвращает класс валидатора ранера функции."""
        return BaseCollectingExportedDataRunnerValidator

    def _prepare_result_class(self) -> type[BaseCollectingExportedDataRunnerResult]:
        """Возвращает класс результата ранера функции."""
        return BaseCollectingExportedDataRunnerResult



# Generated by Django 3.2.24 on 2025-10-17 14:15

from django.db import (
    migrations,
)

from edu_rdm_integration.stages.utils import (
    update_foreign_key_constraint,
)


def apply_fk_updates(apps, schema_editor):
    """
    Применяет обновления внешних ключей для конкретных моделей.

    Здесь задаются таблицы и поведение при удалении записей,
    чтобы синхронизировать фактическое состояние БД с логикой моделей.
    """
    RDMExportingDataSubStage = apps.get_model('edu_rdm_integration_export_data_stage', 'RDMExportingDataSubStage')
    RDMExportingDataStage = apps.get_model('edu_rdm_integration_export_data_stage', 'RDMExportingDataStage')

    RDMExportingDataSubStageAttachment = apps.get_model(
        'edu_rdm_integration_export_data_stage', 'RDMExportingDataSubStageAttachment'
    )
    # Модель: RDMExportingDataSubStageAttachment
    # Поле exporting_data_sub_stage_id должно иметь ON DELETE CASCADE
    update_foreign_key_constraint(
        table_name=RDMExportingDataSubStageAttachment._meta.db_table,
        field_name='exporting_data_sub_stage_id',
        target_table=RDMExportingDataSubStage._meta.db_table,
        on_delete='SET NULL',
    )

    RDMExportingDataSubStageEntity = apps.get_model(
        'edu_rdm_integration_export_data_stage', 'RDMExportingDataSubStageEntity'
    )
    # Модель: CollectDataCommandProgress
    # Поле exporting_data_sub_stage_id должно иметь ON DELETE CASCADE
    update_foreign_key_constraint(
        table_name=RDMExportingDataSubStageEntity._meta.db_table,
        field_name='exporting_data_sub_stage_id',
        target_table=RDMExportingDataSubStage._meta.db_table,
        on_delete='CASCADE',
    )

    RDMExportingDataCommandProgress = apps.get_model(
        'edu_rdm_integration_export_data_stage', 'RDMExportingDataCommandProgress'
    )
    # Модель: RDMExportingDataCommandProgress
    # Поле stage_id должно иметь ON DELETE SET NULL
    update_foreign_key_constraint(
        table_name=RDMExportingDataCommandProgress._meta.db_table,
        field_name='stage_id',
        target_table=RDMExportingDataStage._meta.db_table,
        on_delete='SET NULL',
    )

    # Таблица: rdm_exporting_data_command_progress
    # Поле stage_id должно иметь ON DELETE SET NULL
    update_foreign_key_constraint(
        table_name='rdm_exporting_data_command_progress',
        field_name="stage_id",
        target_table=RDMExportingDataStage._meta.db_table,
        on_delete="SET NULL",
    )


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration_export_data_stage', '0003_alter_rdmexportingdatasubstageattachment_exporting_data_sub_stage'),
    ]

    operations = [
        migrations.RunPython(apply_fk_updates, reverse_code=migrations.RunPython.noop),
    ]


import os
from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    TYPE_CHECKING,
    Iterable,
    Optional,
)

from django.conf import (
    settings,
)
from django.utils import (
    timezone,
)

from edu_function_tools.managers import (
    EduRunnerManager,
)
from educommon import (
    logger,
)
from educommon.utils.date import (
    get_today_min_datetime,
)
from m3_db_utils.consts import (
    DEFAULT_ORDER_NUMBER,
)
from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration.core.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
)
from edu_rdm_integration.core.operations import (
    BaseOperationData,
)
from edu_rdm_integration.core.signals import (
    manager_created,
)
from edu_rdm_integration.core.storages import (
    RegionalDataMartEntityStorage,
)
from edu_rdm_integration.rdm_models.models import (
    RDMModelEnum,
)
from edu_rdm_integration.stages.collect_data.helpers import (
    get_collecting_managers_max_period_ended_dates,
)
from edu_rdm_integration.stages.collect_data.models import (
    RDMCollectingDataStage,
    RDMCollectingDataStageStatus,
)


if TYPE_CHECKING:
    from educommon.audit_log.models import (
        AuditLog,
    )

    from edu_rdm_integration.stages.collect_data.generators import (
        BaseEduLogGenerator,
    )


class BaseCollectModelsData(BaseOperationData):
    """Базовый класс сбора данных моделей РВД."""

    def __init__(
        self,
        models: Iterable[str],
        logs_period_started_at=datetime.combine(date.today(), time.min),
        logs_period_ended_at=datetime.combine(date.today(), time.min),
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Если модели не указаны, берется значение по умолчанию - все модели:
        model_enum_data = RDMModelEnum.get_enum_data()
        models = models if models else model_enum_data.keys()
        self.models: list[ModelEnumValue] = [v for k, v in model_enum_data.items() if k in models]

        self.logs_period_started_at = logs_period_started_at
        self.logs_period_ended_at = logs_period_ended_at

        # Классы менеджеров Функций, которые должны быть запущены для сбора данных моделей РВД
        self._collecting_data_managers: dict[str, type[EduRunnerManager]] = {}

        # Результаты работы Функций сбора данных моделей РВД
        self._collecting_data_results = []

    @property
    def _log_file_path(self) -> str:
        """Путь до лог файла."""
        return os.path.join(settings.MEDIA_ROOT, settings.RDM_COLLECT_LOG_DIR, f'{self.command_id}.log')

    def _has_stage_created_or_in_progress(self, manager_id: str, model: str) -> bool:
        """Проверяет есть ли готовый к работе stage или в работе для данной модели.

        В общем случае разрешается параллельный сбор данных одной модели.
        """
        return False

    def _find_collecting_models_data_managers(self):
        """Поиск менеджеров Функций, которые должны быть запущены для сбора данных моделей РВД."""
        logger.info('collecting models data managers..')

        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        collecting_models_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )

        models = filter(lambda model: model.order_number != DEFAULT_ORDER_NUMBER, self.models)

        for model_enum in models:
            manager_class = collecting_models_data_managers_map.get(model_enum.key)

            if manager_class and not self._has_stage_created_or_in_progress(manager_class.uuid, model_enum.key):
                self._collecting_data_managers[model_enum.key] = manager_class

        logger.info('collecting models data managers finished')

    def _collect_models_data(self, *args, logs: Optional[dict[str, list['AuditLog']]] = None, **kwargs):
        """Запуск Функций по формированию данных моделей РВД из логов."""
        logger.info('collect models data..')

        kwargs['logs_period_started_at'] = self.logs_period_started_at
        kwargs['logs_period_ended_at'] = self.logs_period_ended_at

        for model_key, manager_class in self._collecting_data_managers.items():
            model_logs = logs.get(model_key) if logs else None
            manager = manager_class(*args, logs=model_logs, command_id=self.command_id, **kwargs)

            if self.command_id:
                # Подается сигнал, что менеджер создан:
                manager_created.send(sender=manager, command_id=self.command_id)

            manager.run()

            self._collecting_data_results.append(manager.result)

        logger.info('collecting models data finished.')

    def _collect(self):
        """Собирает данные моделей РВД."""
        logger.info(f'start collecting data of models - {", ".join([model.key for model in self.models])}..')

        self._find_collecting_models_data_managers()
        self._collect_models_data()

        logger.info('collecting data of models is finished.')

    def collect(self):
        """Запускает сбор данных."""
        try:
            self._collect()
        except Exception as err:
            logger.exception(err)
            raise err
        finally:
            self._remove_file_handler()


class BaseCollectModelsDataByGeneratingLogs(BaseCollectModelsData):
    """Сбор данных моделей РВД на основе существующих в БД данных моделей ЭШ.

    Можно регулировать, для каких моделей должен быть произведен сбор данных, и период, за который должны
    быть собраны логи. Логи формируются в процессе выполнения команды при помощи генератора логов
    EduSchoolLogGenerator для указанной модели.
    """

    def __init__(
        self,
        models: Iterable[str],
        logs_period_started_at=datetime.combine(date.today(), time.min),
        logs_period_ended_at=datetime.combine(date.today(), time.min),
        institute_ids=(),
        **kwargs,
    ):
        super().__init__(models, logs_period_started_at, logs_period_ended_at, **kwargs)

        # Учебные заведения, для которых производится выгрузка
        self.institute_ids = institute_ids
        # Генератор логов
        self.log_generator = self._prepare_log_generator()

    def _prepare_log_generator(self) -> 'BaseEduLogGenerator':
        """Возвращает генератор логов."""
        raise NotImplementedError

    def _generate_logs(self) -> list[tuple[dict[str, list['AuditLog']], datetime, datetime]]:
        """Генерация логов.

        Осуществляет генерацию логов по уже существующим записям в базе данных. В качестве параметров указываются
        начало и конец периода сбора логов. Генерация логов производится только для указанных моделей.
        """
        temp_logs: dict[str, list['AuditLog']] = {}

        for model in self.models:
            logs = self.log_generator.generate(
                model=model,
                logs_period_started_at=self.logs_period_started_at,
                logs_period_ended_at=self.logs_period_ended_at,
                institute_ids=self.institute_ids,
            )

            temp_logs[model.key] = logs

        return [(temp_logs, self.logs_period_started_at, self.logs_period_ended_at)]

    def _collect(self):
        """Собирает данные моделей РВД."""
        logger.info(f'start collecting data of models - {", ".join([model.key for model in self.models])}..')
        self._find_collecting_models_data_managers()

        temp_kwargs = {}

        for logs, logs_period_started_at, logs_period_ended_at in self._generate_logs():
            temp_kwargs['logs_period_started_at'] = logs_period_started_at
            temp_kwargs['logs_period_ended_at'] = logs_period_ended_at

            self._collect_models_data(logs=logs, **temp_kwargs)

        logger.info('collecting data of models is finished.')


class BaseCollectLatestModelsData(BaseCollectModelsData):
    """Сбор данных моделей РВД на основе логов за период с последней сборки до указанной даты."""

    def __init__(self, *args, use_times_limit: bool = False, **kwargs):
        super().__init__(*args, **kwargs)

        # Если этот параметр не указан - то высчитываем временные рамки по менеджеру, не учитывая переданные
        # logs_period_started_at и  logs_period_ended_at
        self.use_times_limit = use_times_limit

    def _has_stage_created_or_in_progress(self, manager_id: str, model: str) -> bool:
        """Проверяет есть ли готовый к работе stage или в работе для данной модели."""
        stage_created_or_in_progress = RDMCollectingDataStage.objects.filter(
            manager_id=manager_id,
            status_id__in=(RDMCollectingDataStageStatus.CREATED.key, RDMCollectingDataStageStatus.IN_PROGRESS.key),
        ).exists()

        if stage_created_or_in_progress:
            logger.info(f'model {model} is skipped because it is already created or in progress!')

        return stage_created_or_in_progress

    def _collect_models_data(self, *args, logs: Optional[dict[str, list['AuditLog']]] = None, **kwargs) -> None:
        """Запуск Функций по формированию данных из логов для дальнейшей выгрузки."""
        logger.info('collect models data..')

        managers_last_period_ended_at = get_collecting_managers_max_period_ended_dates(
            self._collecting_data_managers.values()
        )

        for model_key, manager_class in self._collecting_data_managers.items():
            model_logs = logs.get(model_key) if logs else None
            manager_last_period_ended_at = managers_last_period_ended_at.get(manager_class.uuid)

            kwargs['logs_period_started_at'] = (
                self.logs_period_started_at
                if self.use_times_limit
                else (manager_last_period_ended_at or get_today_min_datetime())
            )
            kwargs['logs_period_ended_at'] = self.logs_period_ended_at if self.use_times_limit else timezone.now()

            manager = manager_class(*args, logs=model_logs, **kwargs)

            if self.command_id:
                # Подается сигнал, что менеджер создан:
                manager_created.send(sender=manager, command_id=self.command_id)

            manager.run()

            self._collecting_data_results.append(manager.result)

        logger.info('collecting models data finished.')


class CollectModelsData(BaseCollectModelsData):
    """Сбор данных моделей РВД за указанных период по существующим логам."""
